// SPDX-License-Identifier: MIT

//! Dual-protocol server implementation.
//! Heavily based on https://github.com/daxpedda/axum-server-dual-protocol
//!
//! See [`bind_dual_protocol()`] and [`DualProtocolAcceptor`].

use std::fmt::Debug;
use std::future::Future;
use std::io::ErrorKind;
use std::net::SocketAddr;
use std::pin::Pin;
use std::task::{Context, Poll};
use std::{io, slice};

use axum::http::{Request, Response};
use axum_server::accept::Accept;
use axum_server::Server;
use hyper::server::conn::AddrStream;
use hyper::service::Service as HyperService;
use hyper::Body;
use pin_project::pin_project;
use tokio::io::ReadBuf;
use tokio_util::either::Either as TokioEither;

use crate::server::http_either::Either as BodyEither;
use crate::server::tls::{TlsAcceptor, TlsStream};
use crate::server::upgrade_http::UpgradeHttp;

/// Create a [`Server`] that will bind to the provided address, accepting both
/// HTTP and HTTPS on the same port.
#[must_use]
pub fn bind_dual_protocol(
    address: SocketAddr,
    tls_acceptor: TlsAcceptor,
) -> Server<DualProtocolAcceptor> {
    let acceptor = DualProtocolAcceptor::new(tls_acceptor);

    Server::bind(address).acceptor(acceptor)
}

/// The protocol used by this connection. See
/// [`Request::extensions()`](Request::extensions()).
#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
pub enum Protocol {
    /// This connection is encrypted with TLS.
    Tls,
    /// This connection is unencrypted.
    Plain,
}

/// Simultaneous HTTP and HTTPS [`Accept`]or.
#[derive(Debug, Clone)]
pub struct DualProtocolAcceptor {
    /// [`TlsAcceptor`] re-used to handle HTTPS requests.
    rustls: TlsAcceptor,
}

impl DualProtocolAcceptor {
    /// Create a new [`DualProtocolAcceptor`].
    #[must_use]
    pub fn new(tls_acceptor: TlsAcceptor) -> Self {
        Self {
            rustls: tls_acceptor,
        }
    }
}

impl<Service> Accept<AddrStream, Service> for DualProtocolAcceptor {
    type Stream = TokioEither<TlsStream, AddrStream>;
    type Service = DualProtocolService<Service>;
    type Future = DualProtocolAcceptorFuture<Service>;

    fn accept(&self, stream: AddrStream, service: Service) -> Self::Future {
        let service = DualProtocolServiceBuilder::new_upgrade(service);

        DualProtocolAcceptorFuture::new(stream, service, self.rustls.clone())
    }
}

/// [`Future`](Accept::Future) type for [`DualProtocolAcceptor`].
#[derive(Debug)]
#[pin_project(project = DualProtocolAcceptorFutureProj)]
pub struct DualProtocolAcceptorFuture<Service>(
    /// State. `enum` variants can't be private, so this solution was used to
    /// hide implementation details.
    #[pin]
    FutureState<Service>,
);

/// State of accepting a new request for [`DualProtocolAcceptorFuture`].
#[derive(Debug)]
#[pin_project(project = FutuereStateProj)]
enum FutureState<Service> {
    /// Peeking state, still trying to determine if the incoming request is HTTP
    /// or HTTPS.
    Peek(Option<PeekState<Service>>),
    /// HTTPS state, it was determined that the incoming request is HTTPS, now
    /// the [`TlsAcceptor`] has to be polled to completion.
    Https(#[pin] <TlsAcceptor as Accept<AddrStream, DualProtocolService<Service>>>::Future),
}

/// Data necessary to peek and proceed to the next state.
#[derive(Debug)]
struct PeekState<Service> {
    /// Transport.
    stream: AddrStream,
    /// User-provided [`Service`](hyper::service::Service)
    service: DualProtocolServiceBuilder<Service>,
    /// Used to proceed to the [`Https`](FutureState::Https) state if
    /// necessary.
    rustls: TlsAcceptor,
}

impl<Service> DualProtocolAcceptorFuture<Service> {
    /// Create a new [`DualProtocolAcceptorFuture`] in the
    /// [`Peek`](FutureState::Peek) state.
    const fn new(
        stream: AddrStream,
        service: DualProtocolServiceBuilder<Service>,
        rustls: TlsAcceptor,
    ) -> Self {
        Self(FutureState::Peek(Some(PeekState {
            stream,
            service,
            rustls,
        })))
    }
}

impl<Service> DualProtocolAcceptorFutureProj<'_, Service> {
    /// Proceed to the [`Https`](FutureState::Https) state.
    fn upgrade(
        &mut self,
        future: <TlsAcceptor as Accept<AddrStream, DualProtocolService<Service>>>::Future,
    ) {
        self.0.set(FutureState::Https(future));
    }
}

impl<Service> Future for DualProtocolAcceptorFuture<Service> {
    type Output = io::Result<(
        TokioEither<TlsStream, AddrStream>,
        DualProtocolService<Service>,
    )>;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        let mut this = self.project();

        // After successfully peeking, continue without unnecessary yielding.
        loop {
            match this.0.as_mut().project() {
                FutuereStateProj::Peek(inner) => {
                    let peek = inner.as_mut().expect("polled again after `Poll::Ready`");

                    let mut byte = 0;
                    let mut buffer = ReadBuf::new(slice::from_mut(&mut byte));

                    match peek.stream.poll_peek(cx, &mut buffer) {
                        // If `MSG_PEEK` returns `0`, the socket was closed.
                        Poll::Ready(Ok(0)) => {
                            return Poll::Ready(Err(ErrorKind::UnexpectedEof.into()))
                        }
                        Poll::Ready(Ok(_)) => {
                            let PeekState {
                                stream,
                                service,
                                rustls,
                            } = inner.take().expect("`inner` was already consumed");

                            // The first byte in the TLS protocol is always `0x16`.
                            if byte == 0x16 {
                                this.upgrade(rustls.accept(stream, service.build(Protocol::Tls)));
                            } else {
                                return Poll::Ready(Ok((
                                    TokioEither::Right(stream),
                                    service.build(Protocol::Plain),
                                )));
                            }
                        }
                        Poll::Ready(Err(err)) => return Poll::Ready(Err(err)),
                        Poll::Pending => return Poll::Pending,
                    }
                }
                FutuereStateProj::Https(future) => {
                    return future
                        .poll(cx)
                        .map_ok(|(stream, service)| (TokioEither::Left(stream), service))
                }
            }
        }
    }
}

/// Hold the user-supplied app until the protocol type is determined.
#[derive(Debug)]
struct DualProtocolServiceBuilder<Service>(UpgradeHttp<Service>);

/// [`Service`](HyperService) wrapping user-supplied app to apply global
/// [`Layer`](tower_layer::Layer)s according to configuration.
#[derive(Debug)]
pub struct DualProtocolService<Service> {
    /// The user-supplied [`Service`](HyperService).
    service: UpgradeHttp<Service>,
    /// The protocol this connection is using.
    protocol: Protocol,
}

impl<Service> DualProtocolServiceBuilder<Service> {
    /// Create a [`DualProtocolService`] in the
    /// [`Upgrade`](UpgradeHttp::Upgrade) state.
    const fn new_upgrade(service: Service) -> Self {
        Self(UpgradeHttp::new(service))
    }

    /// Create a [`DualProtocolService`] when the protocol is established.
    #[allow(clippy::missing_const_for_fn)]
    fn build(self, protocol: Protocol) -> DualProtocolService<Service> {
        DualProtocolService {
            service: self.0,
            protocol,
        }
    }
}

impl<Service, RequestBody, ResponseBody> HyperService<Request<RequestBody>>
    for DualProtocolService<Service>
where
    Service: HyperService<Request<RequestBody>, Response = Response<ResponseBody>>,
{
    type Response = Response<BodyEither<ResponseBody, Body>>;
    type Error = Service::Error;
    type Future = DualProtocolServiceFuture<Service, RequestBody, ResponseBody>;

    fn poll_ready(&mut self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
        self.service.poll_ready(cx)
    }

    fn call(&mut self, mut request: Request<RequestBody>) -> Self::Future {
        let _ = request.extensions_mut().insert(self.protocol);

        DualProtocolServiceFuture::new_upgrade(self.service.call(request))
    }
}

/// [`Future`](HyperService::Future) type for [`DualProtocolService`].
#[pin_project]
pub struct DualProtocolServiceFuture<Service, RequestBody, ResponseBody>(
    #[pin] <UpgradeHttp<Service> as HyperService<Request<RequestBody>>>::Future,
)
where
    Service: HyperService<Request<RequestBody>, Response = Response<ResponseBody>>;

impl<Service, RequestBody, ResponseBody>
    DualProtocolServiceFuture<Service, RequestBody, ResponseBody>
where
    Service: HyperService<Request<RequestBody>, Response = Response<ResponseBody>>,
{
    const fn new_upgrade(
        future: <UpgradeHttp<Service> as HyperService<Request<RequestBody>>>::Future,
    ) -> Self {
        Self(future)
    }
}

impl<Service, RequestBody, ResponseBody> Future
    for DualProtocolServiceFuture<Service, RequestBody, ResponseBody>
where
    Service: HyperService<Request<RequestBody>, Response = Response<ResponseBody>>,
{
    type Output = Result<Response<BodyEither<ResponseBody, Body>>, Service::Error>;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        self.project().0.poll(cx)
    }
}
