// SPDX-License-Identifier: MIT

//! HTTP to HTTPS upgrade implementation.
//! Heavily based on https://github.com/daxpedda/axum-server-dual-protocol
//!
//! See [`UpgradeHttp`].

use std::fmt::{self, Debug, Formatter};
use std::future::Future;
use std::pin::Pin;
use std::task::{Context, Poll};

use axum::http::header::{HOST, LOCATION, UPGRADE};
use axum::http::uri::{Authority, Scheme};
use axum::http::{HeaderValue, Request, Response, StatusCode, Uri};
use hyper::service::Service as HyperService;
use hyper::Body;
use pin_project::pin_project;
use tower_layer::Layer;

use crate::server::dual::Protocol;
use crate::server::http_either::Either;

/// [`Layer`] upgrading HTTP requests to HTTPS.
///
/// See [`UpgradeHttp`] for more details.
#[derive(Clone, Copy, Debug)]
pub struct UpgradeHttpLayer;

impl<Service> Layer<Service> for UpgradeHttpLayer {
    type Service = UpgradeHttp<Service>;

    fn layer(&self, service: Service) -> Self::Service {
        UpgradeHttp::new(service)
    }
}

/// [`Service`](HyperService) upgrading HTTP requests to HTTPS by using a
/// [301 "Moved Permanently"](https://tools.ietf.org/html/rfc7231#section-6.4.2)
/// status code.
///
/// Note that this [`Service`](HyperService) always redirects with the given
/// path and query. Depending on how you apply this [`Service`](HyperService) it
/// will redirect even in the case of a resulting 404 "Not Found" status code at
/// the destination.
#[derive(Clone, Debug)]
pub struct UpgradeHttp<Service> {
    /// Wrapped user-proided [`Service`](HyperService).
    service: Service,
}

impl<Service> UpgradeHttp<Service> {
    /// Creates a new [`UpgradeHttp`].
    pub const fn new(service: Service) -> Self {
        Self { service }
    }
}

impl<Service, RequestBody, ResponseBody> HyperService<Request<RequestBody>> for UpgradeHttp<Service>
where
    Service: HyperService<Request<RequestBody>, Response = Response<ResponseBody>>,
{
    type Response = Response<Either<ResponseBody, Body>>;
    type Error = Service::Error;
    type Future = UpgradeHttpFuture<Service, Request<RequestBody>>;

    fn poll_ready(&mut self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
        self.service.poll_ready(cx)
    }

    fn call(&mut self, request: Request<RequestBody>) -> Self::Future {
        match request
            .extensions()
            .get::<Protocol>()
            .expect("`Protocol` should always be set by `DualProtocolService`")
        {
            Protocol::Tls => UpgradeHttpFuture::new_service(self.service.call(request)),
            Protocol::Plain => {
                let response = Response::builder();

                let response = if let Some((authority, scheme)) = extract_authority(&request)
                    .and_then(|authority| {
                        let uri = request.uri();

                        // Depending on the scheme we need a different scheme to redirect to.

                        // WebSocket handshakes often don't send a scheme, so we check the "Upgrade"
                        // header as well.
                        if uri.scheme_str() == Some("ws")
                            || request.headers().get(UPGRADE)
                                == Some(&HeaderValue::from_static("websocket"))
                        {
                            Some((
                                authority,
                                Scheme::try_from("wss").expect("ASCII string is valid"),
                            ))
                        }
                        // HTTP requests often don't send a scheme.
                        else if uri.scheme() == Some(&Scheme::HTTP) || uri.scheme_str().is_none()
                        {
                            Some((authority, Scheme::HTTPS))
                        }
                        // Unknown scheme, abort.
                        else {
                            None
                        }
                    }) {
                    // Build URI to redirect to.
                    let mut uri = Uri::builder().scheme(scheme).authority(authority);

                    if let Some(path_and_query) = request.uri().path_and_query() {
                        uri = uri.path_and_query(path_and_query.clone());
                    }

                    let uri = uri.build().expect("invalid path and query");

                    response
                        .status(StatusCode::MOVED_PERMANENTLY)
                        .header(LOCATION, uri.to_string())
                } else {
                    // If we can't extract the host or have an unknown scheme, tell the client there
                    // is something wrong with their request.
                    response.status(StatusCode::BAD_REQUEST)
                }
                .body(Body::empty())
                .expect("invalid header or body");

                UpgradeHttpFuture::new_upgrade(response)
            }
        }
    }
}

/// [`Future`](HyperService::Future) type for [`UpgradeHttp`].
#[pin_project]
pub struct UpgradeHttpFuture<Service, Request>(#[pin] FutureServe<Service, Request>)
where
    Service: HyperService<Request>;

/// Holds [`Future`] to serve for [`UpgradeHttpFuture`].
#[derive(Debug)]
#[pin_project(project = UpgradeHttpFutureProj)]
enum FutureServe<Service, Request>
where
    Service: HyperService<Request>,
{
    /// The request was using the HTTPS protocol, so we
    /// will pass-through the wrapped [`Service`](HyperService).
    Service(#[pin] Service::Future),
    /// The request was using the HTTP protocol, so we
    /// will upgrade the connection.
    Upgrade(Option<Response<Body>>),
}

// Rust can't figure out the correct bounds.
impl<Service, Request> Debug for UpgradeHttpFuture<Service, Request>
where
    Service: HyperService<Request>,
    FutureServe<Service, Request>: Debug,
{
    fn fmt(&self, formatter: &mut Formatter<'_>) -> fmt::Result {
        formatter
            .debug_tuple("UpgradeHttpFuture")
            .field(&self.0)
            .finish()
    }
}

impl<Service, Request> UpgradeHttpFuture<Service, Request>
where
    Service: HyperService<Request>,
{
    /// Create a [`UpgradeHttpFuture`] in the [`Service`](FutureServe::Service)
    /// state.
    const fn new_service(future: Service::Future) -> Self {
        Self(FutureServe::Service(future))
    }

    /// Create a [`UpgradeHttpFuture`] in the [`Upgrade`](FutureServe::Upgrade)
    /// state.
    const fn new_upgrade(response: Response<Body>) -> Self {
        Self(FutureServe::Upgrade(Some(response)))
    }
}

impl<Service, Request, ResponseBody> Future for UpgradeHttpFuture<Service, Request>
where
    Service: HyperService<Request, Response = Response<ResponseBody>>,
{
    type Output = Result<Response<Either<ResponseBody, Body>>, Service::Error>;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        match self.project().0.project() {
            UpgradeHttpFutureProj::Service(future) => {
                future.poll(cx).map_ok(|result| result.map(Either::Left))
            }
            UpgradeHttpFutureProj::Upgrade(response) => Poll::Ready(Ok(response
                .take()
                .expect("polled again after `Poll::Ready`")
                .map(Either::Right))),
        }
    }
}

/// Extracts the host from a request, converting it to an [`Authority`].
fn extract_authority<Body>(request: &Request<Body>) -> Option<Authority> {
    /// `X-Forwarded-Host` header string.
    const X_FORWARDED_HOST: &str = "x-forwarded-host";

    let headers = request.headers();

    headers
        .get(X_FORWARDED_HOST)
        .or_else(|| headers.get(HOST))
        .and_then(|header| header.to_str().ok())
        .or_else(|| request.uri().host())
        .and_then(|host| Authority::try_from(host).ok())
}
