use axum::async_trait;
use parking_lot::{Mutex, MutexGuard};
use std::{
    collections::VecDeque,
    convert::Infallible,
    sync::Arc,
    time::{Duration, Instant},
};
use tower_sessions::{
    session::{Id, Session},
    SessionStore,
};

#[derive(Debug)]
struct StoreItem {
    id: Id,
    session: Session,
    stored_date: Instant,
}

#[derive(Clone, Debug)]
pub struct MemoryStore {
    store: Arc<Mutex<VecDeque<StoreItem>>>,
    lifetime: Duration,
}

impl MemoryStore {
    pub fn new(lifetime: Duration) -> Self {
        MemoryStore {
            store: Mutex::new(VecDeque::new()).into(),
            lifetime,
        }
    }

    fn is_active(&self, item: &StoreItem) -> bool {
        item.stored_date.elapsed() < self.lifetime
    }

    fn purge_old(&self, queue: &mut MutexGuard<VecDeque<StoreItem>>) {
        while queue.get(0).is_some_and(|item| !self.is_active(item)) {
            queue.pop_front();
        }
    }
}

#[async_trait]
impl SessionStore for MemoryStore {
    type Error = Infallible;

    async fn save(&self, session: &Session) -> Result<(), Self::Error> {
        let mut store = self.store.lock();
        self.purge_old(&mut store);
        store.push_back(StoreItem {
            id: *session.id(),
            session: session.clone(),
            stored_date: Instant::now(),
        });
        Ok(())
    }

    async fn load(&self, session_id: &Id) -> Result<Option<Session>, Self::Error> {
        let mut store = self.store.lock();
        self.purge_old(&mut store);
        for item in store.iter() {
            if &item.id == session_id {
                return Ok(Some(item.session.clone()));
            }
        }
        Ok(None)
    }

    async fn delete(&self, session_id: &Id) -> Result<(), Self::Error> {
        let mut store = self.store.lock();
        self.purge_old(&mut store);
        for (idx, item) in store.iter().enumerate() {
            if &item.id == session_id {
                store.remove(idx);
                break;
            }
        }
        Ok(())
    }
}
